from json import JSONDecodeError
import logging
from typing import Dict

import requests
import requests.packages

from pyostrap.exceptions import BiostrapApiException
from pyostrap.models import Result


class RestAdapter:
    def __init__(
        self,
        api_key: str,
        hostname: str = "api-beta.biostrap.com",
        ver: str = "v1",
        ssl_verify: bool = True,
        logger: logging.Logger = None,
    ):
        """
        :param api_key: string used for authentication
        :param hostname: Normally, api-beta.biostrap.com
        :param ver: always v1
        :param ssl_verify: Normally set to True, but if having SSL/TLS cert validation issues, can turn off with False
        :param logger: (optional) If your app has a logger, pass it in here
        """
        self.url = f"https://{hostname}/{ver}/"
        self._api_key = api_key
        self._ssl_verify = ssl_verify
        self._logger = logger or logging.getLogger(__name__)
        if not ssl_verify:
            requests.packages.urllib3.disable_warnings()

    def _do(
        self, http_method: str, endpoint: str, ep_params: Dict = None, data: Dict = None
    ) -> Result:
        full_url = self.url + endpoint
        headers = {"Authorization": f"APIKey {self._api_key}"}
        log_line_pre = f"method={http_method}, url={full_url}, params={ep_params}"
        # Log HTTP params and perform an HTTP request, catching and re-raising any exceptions
        try:
            self._logger.debug(msg=log_line_pre)
            response = requests.request(
                method=http_method,
                url=full_url,
                verify=self._ssl_verify,
                headers=headers,
                params=ep_params,
                json=data,
            )
        except requests.exceptions.RequestException as e:
            self._logger.error(msg=(str(e)))
            raise BiostrapApiException("Request failed") from e
        # Deserialize JSON output to Python object, or return failed Result on exception
        try:
            data_out = response.json()
        except (ValueError, JSONDecodeError) as e:
            log_line = (
                f"{log_line_pre}, success={False}, status_code={None}, message={e}"
            )
            self._logger.error(log_line)
            raise BiostrapApiException("Bad JSON in response") from e

        is_success = 200 <= response.status_code <= 299
        log_line = f"{log_line_pre}, success={is_success}, status_code={response.status_code}, message={response.reason}"

        if is_success:
            self._logger.debug(msg=log_line)
            return Result(response.status_code, message=response.reason, data=data_out)
        self._logger.error(msg=log_line)
        raise BiostrapApiException(f"{response.status_code}: {response.reason}")

    def get(self, endpoint: str, ep_params: Dict = None) -> Result:
        return self._do(http_method="GET", endpoint=endpoint, ep_params=ep_params)

    def post(self, endpoint: str, ep_params: Dict = None, data: Dict = None) -> Result:
        return self._do(
            http_method="POST", endpoint=endpoint, ep_params=ep_params, data=data
        )



from datetime import date, datetime
import logging
from typing import List

from pyostrap.models import (
    CalorieDetailsGranular,
    DeviceInfo,
    JobStatus,
    LockStatus,
    Pagination,
    Scores,
    User,
    Users,
)
from pyostrap.rest_adapter import RestAdapter
from pyostrap.util import get_rfc3339_str


class BiostrapApi:
    def __init__(
        self,
        api_key: str,
        hostname: str = "api-beta.biostrap.com",
        ver: str = "v1",
        ssl_verify: bool = True,
        logger: logging.Logger = None,
    ):
        self._rest_adapter = RestAdapter(api_key, hostname, ver, ssl_verify, logger)

    # Calories
    def get_calorie_details_granular(
        self,
        user_id: str,
        date: date,
        granularity: str,
        user_timezone_offset_in_mins: int = 0,
    ):
        ep_params = {
            "user_id": user_id,
            "user_timezone_offset_in_mins": user_timezone_offset_in_mins,
            "date": date.strftime("%Y-%m-%d"),
            "granularity": granularity,
        }
        result = self._rest_adapter.get(endpoint="calorie/details", ep_params=ep_params)
        return CalorieDetailsGranular(**result.data)

    # Device Information
    def get_device_info(self, user_id: str) -> List[DeviceInfo]:
        ep_params = {"user_id": user_id}
        result = self._rest_adapter.get(endpoint="device-info", ep_params=ep_params)
        return [DeviceInfo(**device) for device in result.data["devices"]]

    # Organizations
    def download_raw_data(
        self,
        target_email: str,
        start_time: datetime,
        end_time: datetime,
        user_ids: List[str],
        data_to_download: List[str],
        output_file_formats: List[str],
        anonymize_ids: bool = False,
    ) -> str:
        json_body = {
            "target_email": target_email,
            "start_time": get_rfc3339_str(start_time),
            "end_time": get_rfc3339_str(end_time),
            "user_ids": user_ids,
            "data_to_download": data_to_download,
            "output_file_formats": output_file_formats,
            "anonymize_ids": anonymize_ids,
        }
        result = self._rest_adapter.post(
            endpoint="organizations/data-download/raw/send-request", data=json_body
        )
        return result.data["job_id"]

    def get_job_status(self, job_id: str) -> JobStatus:
        ep_params = {"job_id": job_id}
        result = self._rest_adapter.get(
            endpoint="organizations/job-status", ep_params=ep_params
        )
        return JobStatus(**result.data["data"])

    def lock_or_unlock_device_to_user(
        self,
        user_id: str,
        device_type: str,
        device_mac_address_or_id_encoded: str,
        operation: str,
    ):
        json_body = {
            "user_id": user_id,
            "device_type": device_type,
            "device_mac_address_or_id_encoded": device_mac_address_or_id_encoded,
            "operation": operation,
        }

        result = self._rest_adapter.post(
            endpoint="organizations/user-device-lock", data=json_body
        )
        return LockStatus(**result.data)

    def get_users(self, page: int, items_per_page: int) -> Users:
        ep_params = {"page": page, "items_per_page": items_per_page}
        result = self._rest_adapter.get(
            endpoint="organizations/users", ep_params=ep_params
        )
        pagination = Pagination(**result.data["pagination"])
        user_list = [User(**raw_user) for raw_user in result.data["users"]]
        return Users(user_list, pagination.page < pagination.available_pages)

    # Scores
    def get_user_scores(self, day: date, user_id: str) -> Scores:
        ep_params = {"date": day.isoformat(), "user_id": user_id}
        result = self._rest_adapter.get(endpoint="scores", ep_params=ep_params)
        return Scores(**result.data)

    # Users
    def get_user(self, user_id: str) -> User:
        ep_params = {"user_id": user_id}
        result = self._rest_adapter.get(endpoint="user", ep_params=ep_params)
        return User(**result.data["data"])



from dataclasses import dataclass
from datetime import datetime
from typing import List, Dict


class Result:
    def __init__(self, status_code: int, message: str = "", data: List[Dict] = None):
        """
        Result returned from low-level RestAdapter
        :param status_code: Standard HTTP Status code
        :param message: Human readable result
        :param data: Python List of Dicts
        """
        self.status_code = int(status_code)
        self.message = str(message)
        self.data = data if data else []


@dataclass
class JobStatus:
    job_id: str
    job_type: str
    latest_status: str
    status_updated_at_ts: int


@dataclass
class Pagination:
    available_pages: int
    items_per_page: int
    page: int
    total_items: int


@dataclass
class Goals:
    steps: int
    sleep: int
    calories: int
    workout: int


class User:
    def __init__(
        self,
        id: str,
        name: str,
        email: str,
        birthday: str,
        gender: str,
        height: float,
        weight: float,
        goals: Dict,
    ):
        self.id = id
        self.name = name
        self.email = email
        self.birthday = datetime.strptime(birthday, "%Y-%m-%d").date()
        self.gender = gender
        self.height = height
        self.weight = weight
        self.goals = Goals(**goals)


class Users:
    def __init__(self, users: List[User], data_left: bool):
        self.users = users
        self.data_left = data_left

    def __iter__(self):
        for user in self.users:
            yield user


@dataclass
class DeviceInfo:
    last_data_uploaded_at_ts: int
    last_updated_at_tz_offset_mins: int
    type: str
    battery_percentage: int = 0
    last_battery_info_updated_at_ts: int = 0


@dataclass
class ActivityScore:
    avg: int
    goal: int
    processing: bool
    value: int


@dataclass
class RecoveryScore:
    avg: int
    message: str
    processing: bool
    stage: str
    value: int


@dataclass
class SleepScore:
    avg: int
    duration_secs: int
    goal: int
    processing: bool
    value: int


class Scores:
    def __init__(self, activity: Dict, recovery: Dict, sleep: Dict):
        self.activity_score = ActivityScore(**activity)
        self.recovery_score = RecoveryScore(**recovery)
        self.sleep_score = SleepScore(**sleep)

@dataclass
class LockStatus:
    status: str
    status_message: str

class Timepoint:
    def __init__(self, time: str, date: str, value: float):
        self.time = None if not time else datetime.fromisoformat(time)
        self.date = None if not self.time else self.time.date()
        self.value = value

class Metric:
    def __init__(self, type: str, name: str, value: float, unit: str, value_is_an_avg: bool, timeseries: List[Dict]):
        self.type = type
        self.name = name
        self.value = value
        self.unit = unit
        self.value_is_an_avg = value_is_an_avg
        self.timeseries = [Timepoint(**timepoint) for timepoint in timeseries]

class CalorieDetailsGranular:
    def __init__(self, date: str, granularity: str, daily_calories_goal: int, calories_goal_achieved_percentage: int, metrics: List[Dict]):
        self.date = datetime.strptime(date, "%Y-%m-%d").date()
        self.granularity = granularity
        self.daily_calories_goal = daily_calories_goal
        self.calories_goal_achieved_percentage = calories_goal_achieved_percentage
        self.metrics = [Metric(**metric) for metric in metrics]